<!---
Example for expanded and simplified checking on complex variable existence.
This code is provided as is under the Apache Creative Commons License v.3

Creator:
Bilal Soylu
July 2010
--->

<cffunction name="varExists" hint="check whether a variable is accessible, handle structures and arrays. Optionally set defaults and return type.">
	<cfargument name="sInput" type="string" required="Yes" hint="the variable reference to be checked for existence, e.g. URL.MyVar, MyArray[2],MyStruct[T2][3]">
	<cfargument name="bGetVal" type="boolean" default="no" hint="whether the check for existence will return YES (default)/NO , or if the definition exists the value of the variable and if not exists: NO">
	<cfargument name="vDefault" hint="a default value to be returned if the variable is not defined">


	<cfscript>
		var vReturn ="YES"; //assume the variable is defined
		var i=1;
		var Local=StructNew();
		try {
			Local.iStart=Find("[",Arguments.sInput);
			//no square brackets, we can use isdefined eval
			if (Local.iStart IS 0) {
				vReturn = IsDefined("#Arguments.sInput#");  //returns "YES" OR "NO"
				//return value if needed
				if (Arguments.bGetVal AND vReturn) vReturn = Evaluate("#Arguments.sInput#");
			} else {
				//we have a struct or array and have to parse in multiple steps
				//we need string right before open [ and string between [ and ]
				Local.er = Left(Arguments.sInput,Local.iStart-1); //eval root element
				//check initial element outside loop
				if (IsDefined("#Local.er#")) {
					Local.tempVar = Evaluate("#Local.er#");

					for (i=2; i LTE ListLen(Arguments.sInput,"[]"); i=i+1) {
						//WriteOutput("<br>" & Local.er & "<br>");

						//get next part and assign part ref
						Local.sPart = ListGetAt(Arguments.sInput,i,"[]");
						//check whether list element is in brackets
						Local.iBracketCheck = Find("[" & Local.sPart & "]",Arguments.sInput,Local.iStart-1);
					    if (Local.iBracketCheck GT 0) {
							//we are in a bracket expression
							Local.iStart = Local.iBracketCheck;
							Local.er="";
							//strip single and double quotes
							Local.sUsePart = Replace(Local.sPart,"'","","ALL");
							Local.sUsePart = Replace(Local.sUsePart,'"','','ALL');
							//test for struct and array, we compare sUsePart to sPart to make sure that there were single or double quotes when associate arrays were used
							//otherwise the var should not be considered as defined
							if (IsStruct(Local.tempVar) AND Len(Local.sUsePart) NEQ Len(Local.sPart) AND StructKeyExists(Local.tempVar,Local.sUsePart)) {
								//assign to tempvar via associative array notation
								Local.tempVar = Local.tempVar[Local.sUsePart];
							} else if (IsArray(Local.tempVar) AND IsNumeric(Local.sPart) AND ArrayLen(Local.tempVar) GTE Local.sPart) {
								//assign array index to temp var so far
								//arrays may have undefined elements if we use non-sequential index assignments
								if (ArrayIsDefined(Local.tempVar, Local.sPart)) {
									try {
										Local.tempVar = Local.tempVar[Local.sPart];
									} catch (ANY e) {
										vReturn="NO";
										break;
									}
								} else {
									vReturn="NO";
									break;									
								}
							} else {
								//we are not defined
								vReturn="NO";
								break;
							}
						} else {
							Local.iStart = Find(Local.sPart, Arguments.sInput, Local.iStart-1);
							//struct part, use simple test
							Local.er = "Local.tempVar" & Local.sPart;
							//check for isDefined
							if (IsDefined("#Local.er#")) {
								Local.tempVar = Evaluate("#Local.er#");
							} else {
								vReturn="NO";
								break;
							}
						}
					} //for loop
					//return value if needed
					if (Arguments.bGetVal AND vReturn) vReturn = Evaluate("#Arguments.sInput#");
				} else {
					//initial part is not defined
					vReturn="NO";
				};

			} //shortcut eval if not a complex expression
		} catch (ANY e) {
			vReturn="NO"; //error occured have to assume var is not defined
		}
		//return default if we did not find defined element and have a default
		if (IsSimpleValue(vReturn) AND IsBoolean(vReturn) AND vReturn IS false AND IsDefined("Arguments.vDefault")) vReturn =  Arguments.vDefault;

		return vReturn;
	</cfscript>


</cffunction>





	<cfscript>
		//Create test data
		//example complex data store to test against
		sTest = StructNew();
		sTest.Animals= StructNew();
		sTest.Animals.favorites = StructNew();
		sTest.Animals.favorites.color="green";
		sTest.Animals.favorites.aspect="night vision";
		sTest.Animals.favorites.feline="panther";
		sTest.Animals.species = ArrayNew(1);

		//lions
		sTest.Animals.species[1]=StructNew();
		sTest.Animals.species[1].CommonName="lions";
		sTest.Animals.species[1].Name="Panthera leo";
		sTest.Animals.species[1].counts=ArrayNew(1);
		sTest.Animals.species[1].counts[1] = StructNew();
		sTest.Animals.species[1].counts[1].small=100;
		sTest.Animals.species[1].counts[1].medium=122;
		sTest.Animals.species[1].counts[1].large=34;
		sTest.Animals.species[1].counts[2] = 256;
		//walruss
		sTest.Animals.species[2] = StructNew();
		sTest.Animals.species[2].CommonName="walrus";
		sTest.Animals.species[2].Name="Odobenus rosmarus";
		sTest.Animals.species[2].counts=ArrayNew(1);
		sTest.Animals.species[2].counts[1] = StructNew();
		sTest.Animals.species[2].counts[1].small=20;
		sTest.Animals.species[2].counts[1].medium=35;
		sTest.Animals.species[2].counts[1].large=25;
		sTest.Animals.species[2].counts[2] = 80;
		sTest.Animals.species[2].counts[10] = 800; //create a gap in array index 3..9
		//multidimensional array
		sTest.Stats = ArrayNew(2);
		sTest.Stats[1][1]=1;
		sTest.Stats[1][2]=2;
		sTest.Stats[1][3]=3;
		sTest.Stats[2][1]=4;
		sTest.Stats[2][2]=5;
	</cfscript>

	
	<cfoutput>
	
	<h3>Unit Testing varExists() function</h3>
	<hr>
	<br>
	
	<!--- no exceptions (bsoylu 07-18-2010) --->
	<cfset testValue = "sTest.Animals.favorites.color">
	<strong>Standard IsDefined check</strong>: #testValue#<br>
	Exists:#varExists("#testValue#")#<br>
	<!--- simple example use to dump test value if exists --->
	Dump:<cfdump var="#varExists(testValue,true,'not defined')#"> <br>
	<hr noshade>
	
	<!--- associative array for structure followed by standard arrays (bsoylu 07-18-2010) --->
	<cfset testValue = "sTest['Animals'].species[2].counts[2]">
	<strong>Structure as associative array</strong>: #testValue#<br>
	Exists:#varExists("#testValue#")#<br>
	<!--- simple example use to dump test value if exists --->
	Dump:<cfdump var="#varExists(testValue,true,'not defined')#"> <br>
	<hr noshade>

	
	<!--- series of associative arrays attached to normal array (bsoylu 07-18-2010) --->
	<cfset testValue = "sTest.Animals['species'][1]['counts'][2]">
	<strong>Series of associative and normal arrays</strong>: #testValue#<br>
	Exists:#varExists("#testValue#")#<br>
	<!--- simple example use to dump test value if exists --->
	Dump:<cfdump var="#varExists(testValue,true,'not defined')#"> <br>
	<hr noshade>


	<!--- return of complex values (bsoylu 07-18-2010) --->
	<cfset testValue = "sTest.Animals.species[2]">
	<strong>Complex value return</strong>: #testValue#<br>
	Exists:#varExists("#testValue#")#<br>
	<!--- simple example use to dump test value if exists --->
	Dump:<cfdump var="#varExists(testValue,true,'not defined')#"> <br>
	<hr noshade>


	<!--- undefined array index (bsoylu 07-18-2010) --->
	<cfset testValue = "sTest.Animals.species[2].counts[4]">
	<strong>Undefined Array Index</strong>: #testValue#<br>
	Exists:#varExists("#testValue#")#<br>
	<!--- simple example use to dump test value if exists --->
	Dump:<cfdump var="#varExists(testValue,true,'not defined')#"> <br>
	<hr noshade>

	<!--- multidimensional array index (bsoylu 07-18-2010) --->
	<cfset testValue = "sTest.Stats[2][2]">
	<strong>multidimensional Array </strong>: #testValue#<br>
	Exists:#varExists("#testValue#")#<br>
	<!--- simple example use to dump test value if exists --->
	Dump:<cfdump var="#varExists(testValue,true,'not defined')#"> <br>
	<hr noshade>

	<!--- undefined reference (bsoylu 07-18-2010) --->
	<cfset testValue = "sTest.Stats[2][2].inputs">
	<strong>undefined reference </strong>: #testValue#<br>
	Exists:#varExists("#testValue#")#<br>
	<!--- simple example use to dump test value if exists --->
	Dump:<cfdump var="#varExists(testValue,true,'not defined')#"> <br>
	<hr noshade>

	</cfoutput>
	<br>
	<h3>Full Test Structure</h3>
	<hr>
	<cfdump var="#sTest#">